import pytest
from aare import RawFile
import numpy as np

@pytest.mark.withdata
def test_read_rawfile_with_roi(test_data_path):

   with RawFile(test_data_path / "raw/SingleChipROI/Data_master_0.json") as f:
    headers, frames = f.read()

    assert headers.size == 10100
    assert frames.shape == (10100, 256, 256)
    
@pytest.mark.withdata
def test_read_rawfile_quad_eiger_and_compare_to_numpy(test_data_path): 
    
    d0 = test_data_path/'raw/eiger_quad_data/W13_vrpreampscan_m21C_300V_800eV_vthre2000_d0_f0_0.raw'
    d1 = test_data_path/'raw/eiger_quad_data/W13_vrpreampscan_m21C_300V_800eV_vthre2000_d1_f0_0.raw'

    image = np.zeros((512,512), dtype=np.uint32)

    with open(d0) as f:
        raw = np.fromfile(f, dtype=np.uint32, count = 256*512, offset = 20*256*512*4 + 112*21).reshape(256,512)

        image[256:,:] = raw

    with open(d1) as f:
        raw = np.fromfile(f, dtype=np.uint32, count = 256*512, offset = 20*256*512*4 + 112*21).reshape(256,512)
        
        image[0:256,:] = raw[::-1,:]

    with RawFile(test_data_path/'raw/eiger_quad_data/W13_vrpreampscan_m21C_300V_800eV_vthre2000_master_0.json') as f:
        f.seek(20)
        header, image1 = f.read_frame()
    
    assert (image == image1).all()


@pytest.mark.withdata
def test_read_rawfile_eiger_and_compare_to_numpy(test_data_path): 
    d0 = test_data_path/'raw/eiger/Lab6_20500eV_2deg_20240629_d0_f0_7.raw'
    d1 = test_data_path/'raw/eiger/Lab6_20500eV_2deg_20240629_d1_f0_7.raw'
    d2 = test_data_path/'raw/eiger/Lab6_20500eV_2deg_20240629_d2_f0_7.raw'
    d3 = test_data_path/'raw/eiger/Lab6_20500eV_2deg_20240629_d3_f0_7.raw'

    image = np.zeros((512,1024), dtype=np.uint32)

    #TODO why is there no header offset?
    with open(d0) as f:
        raw = np.fromfile(f, dtype=np.uint32, count = 256*512, offset=112).reshape(256,512)

        image[0:256,0:512] = raw[::-1]

    with open(d1) as f:
        raw = np.fromfile(f, dtype=np.uint32, count = 256*512, offset=112).reshape(256,512)
        
        image[0:256,512:] = raw[::-1]

    with open(d2) as f:
        raw = np.fromfile(f, dtype=np.uint32, count = 256*512, offset=112).reshape(256,512)
        
        image[256:,0:512] = raw

    with open(d3) as f:
        raw = np.fromfile(f, dtype=np.uint32, count = 256*512, offset=112).reshape(256,512)
        
        image[256:,512:] = raw

    
    with RawFile(test_data_path/'raw/eiger/Lab6_20500eV_2deg_20240629_master_7.json') as f:
        header, image1 = f.read_frame()

    assert (image == image1).all()




from . import _aare
import numpy as np
from .ScanParameters import ScanParameters

class CtbRawFile(_aare.CtbRawFile):
    """File reader for the CTB raw file format.
    
    Args:  
            fname (pathlib.Path | str): Path to the file to be read.
            chunk_size (int): Number of frames to read at a time. Default is 1.
            transform (function): Function to apply to the data after reading it. 
                The function should take a numpy array of type uint8 and return one
                or several numpy arrays.
    """
    def __init__(self, fname, chunk_size = 1, transform = None):
        super().__init__(fname)
        self._chunk_size = chunk_size
        self._transform = transform


    def read_frame(self, frame_index: int | None = None ) -> tuple:
        """Read one frame from the file and then advance the file pointer.

        .. note::

            Uses the position of the file pointer :py:meth:`~CtbRawFile.tell` to determine
            which frame to read unless frame_index is specified.

        Args:
            frame_index (int): If not None, seek to this frame before reading.

        Returns:
            tuple: header, data

        Raises:
            RuntimeError: If the file is at the end.
        """
        if frame_index is not None:
            self.seek(frame_index)


        header, data = super().read_frame()
        if header.shape == (1,):
            header = header[0]


        if self._transform:
            res = self._transform(data)
            if isinstance(res, tuple):
                return header, *res
            else:
                return header, res
        else:
            return header, data
            
    def read_n(self, n_frames:int) -> tuple:
        """Read several frames from the file.

        .. note::

            Uses the position of the file pointer :py:meth:`~CtbRawFile.tell` to determine
            where to start reading from.

        If the number of frames requested is larger than the number of frames left in the file,
        the function will read the remaining frames. If no frames are left in the file
        a RuntimeError is raised.

        Args:
            n_frames (int): Number of frames to read.

        Returns:
            tuple: header, data

        Raises:
            RuntimeError: If EOF is reached.
        """
        # Calculate the number of frames to actually read
        n_frames = min(n_frames, self.frames_in_file - self.tell())
        if n_frames == 0:
            raise RuntimeError("No frames left in file.")


        # Do the first read to figure out what we have
        tmp_header, tmp_data = self.read_frame()
        
        # Allocate arrays for
        header = np.zeros(n_frames, dtype = tmp_header.dtype)
        data = np.zeros((n_frames, *tmp_data.shape), dtype = tmp_data.dtype)
        
        # Copy the first frame
        header[0] = tmp_header
        data[0] = tmp_data

        # Do the rest of the reading
        for i in range(1, n_frames):
            header[i], data[i] = self.read_frame()

        return header, data

    def read(self) -> tuple:
        """Read the entire file.
        Seeks to the beginning of the file before reading.

        Returns:
            tuple: header, data
        """
        self.seek(0)
        return self.read_n(self.frames_in_file)

    def seek(self, frame_index:int) -> None:
        """Seek to a specific frame in the file.

        Args:
            frame_index (int): Frame position in file to seek to.
        """
        super().seek(frame_index)

    def tell(self) -> int:
        """Return the current frame position in the file.

        Returns:
            int: Frame position in file.
        """
        return super().tell()
    
    @property
    def scan_parameters(self):
        """Return the scan parameters.

        Returns:
            ScanParameters: Scan parameters.
        """
        return ScanParameters(self.master.scan_parameters)

    @property
    def master(self):
        """Return the master file.

        Returns:
            RawMasterFile: Master file.
        """
        return super().master()

    @property
    def image_size_in_bytes(self) -> int:
        """Return the size of the image in bytes.

        Returns:
            int: Size of image in bytes.
        """
        return super().image_size_in_bytes

    def __len__(self) -> int:
        """Return the number of frames in the file.

        Returns:
            int: Number of frames in file.
        """
        return super().frames_in_file
    
    @property
    def frames_in_file(self) -> int:
        """Return the number of frames in the file.

        Returns:
            int: Number of frames in file.
        """
        return super().frames_in_file

    def __enter__(self):
        return self
    
    def __exit__(self, exc_type, exc_value, traceback):
        pass

    def __iter__(self):
        return self
    
    def __next__(self):
        try:
            if self._chunk_size == 1:
                return self.read_frame()
            else:
                return self.read_n(self._chunk_size)
            
                
        except RuntimeError:
            # TODO! find a good way to check that we actually have the right exception
            raise StopIteration





from aare import apply_calibration
import numpy as np
raw = np.zeros((5,10,10), dtype=np.uint16)
pedestal = np.zeros((3,10,10), dtype=np.float32)
calibration = np.ones((3,10,10), dtype=np.float32)
calibrated = apply_calibration(raw, pedestal, calibration,)




import os
from pathlib import Path
import pytest



def pytest_addoption(parser):
    parser.addoption(
        "--with-data", action="store_true", default=False, help="Run tests that require additional data"
    )


def pytest_configure(config):
    config.addinivalue_line("markers", "withdata: mark test as needing image files to run")


def pytest_collection_modifyitems(config, items):
    if config.getoption("--with-data"):
        return
    skip = pytest.mark.skip(reason="need --with-data option to run")
    for item in items:
        if "withdata" in item.keywords:
            item.add_marker(skip)


@pytest.fixture
def test_data_path():
    env_value = os.environ.get("AARE_TEST_DATA")
    if not env_value:
        raise RuntimeError("Environment variable AARE_TEST_DATA is not set or is empty")

    return Path(env_value)






import pytest 
import numpy as np
import boost_histogram as bh
import time
from pathlib import Path
import pickle

from aare import ClusterFile
from conftest import test_data_path

@pytest.mark.withdata
def test_cluster_file(test_data_path): 
    """Test ClusterFile""" 
    f =  ClusterFile(test_data_path / "clust/single_frame_97_clustrers.clust") 
    cv = f.read_clusters(10) #conversion does not work


    assert cv.frame_number == 135
    assert cv.size == 10

    #Known data 
    #frame_number, num_clusters   [135] 97
    #[  1 200] [0 1 2 3 4 5 6 7 8]
    #[  2 201] [ 9 10 11 12 13 14 15 16 17]
    #[  3 202] [18 19 20 21 22 23 24 25 26]
    #[  4 203] [27 28 29 30 31 32 33 34 35]
    #[  5 204] [36 37 38 39 40 41 42 43 44]
    #[  6 205] [45 46 47 48 49 50 51 52 53]
    #[  7 206] [54 55 56 57 58 59 60 61 62]
    #[  8 207] [63 64 65 66 67 68 69 70 71]
    #[  9 208] [72 73 74 75 76 77 78 79 80]
    #[ 10 209] [81 82 83 84 85 86 87 88 89]

    #conversion to numpy array
    arr = np.array(cv, copy = False)
    
    assert arr.size == 10
    for i in range(10):
        assert arr[i]['x'] == i+1

@pytest.mark.withdata
def test_read_clusters_and_fill_histogram(test_data_path): 
    # Create the histogram
    n_bins = 100
    xmin = -100
    xmax = 1e4
    hist_aare = bh.Histogram(bh.axis.Regular(n_bins, xmin, xmax))

    fname = test_data_path / "clust/beam_En700eV_-40deg_300V_10us_d0_f0_100.clust"

    #Read clusters and fill the histogram with pixel values
    with ClusterFile(fname, chunk_size = 10000) as f:
        for clusters in f:
            arr = np.array(clusters, copy = False)
            hist_aare.fill(arr['data'].flat)


    #Load the histogram from the pickle file
    with open(fname.with_suffix('.pkl'), 'rb') as f:
        hist_py = pickle.load(f)
        
    #Compare the two histograms
    assert hist_aare == hist_py


from . import _aare

class ScanParameters(_aare.ScanParameters):
    def __init__(self, s):
        super().__init__(s)
        
    def __iter__(self):
        return [getattr(self, a) for a in ['start', 'stop', 'step']].__iter__()
    
    def __str__(self):
        return f'ScanParameters({self.dac}: {self.start}, {self.stop}, {self.step})'
    
    def __repr__(self):
        return self.__str__()
    
    


import pytest
import numpy as np
from aare import JungfrauDataFile

@pytest.mark.withdata
def test_jfungfrau_dat_read_number_of_frames(test_data_path):
    with JungfrauDataFile(test_data_path / "dat/AldoJF500k_000000.dat") as dat_file:
        assert dat_file.total_frames == 24

    with JungfrauDataFile(test_data_path / "dat/AldoJF250k_000000.dat") as dat_file:
        assert dat_file.total_frames == 53
    
    with JungfrauDataFile(test_data_path / "dat/AldoJF65k_000000.dat") as dat_file:
        assert dat_file.total_frames == 113


@pytest.mark.withdata
def test_jfungfrau_dat_read_number_of_file(test_data_path):
    with JungfrauDataFile(test_data_path / "dat/AldoJF500k_000000.dat") as dat_file:
        assert dat_file.n_files == 4

    with JungfrauDataFile(test_data_path / "dat/AldoJF250k_000000.dat") as dat_file:
        assert dat_file.n_files == 7
    
    with JungfrauDataFile(test_data_path / "dat/AldoJF65k_000000.dat") as dat_file:
        assert dat_file.n_files == 7


@pytest.mark.withdata
def test_read_module(test_data_path):
    """
    Read all frames from the series of .dat files. Compare to canned data in npz format. 
    """

    # Read all frames from the .dat file
    with JungfrauDataFile(test_data_path / "dat/AldoJF500k_000000.dat") as f:
        header, data = f.read()

    #Sanity check
    n_frames = 24
    assert header.size == n_frames
    assert data.shape == (n_frames, 512, 1024)

    # Read reference data using numpy
    with np.load(test_data_path / "dat/AldoJF500k.npz") as f:
        ref_header = f["headers"]
        ref_data = f["frames"]

    # Check that the data is the same
    assert np.all(ref_header == header)
    assert np.all(ref_data == data)

@pytest.mark.withdata
def test_read_half_module(test_data_path):

    # Read all frames from the .dat file
    with JungfrauDataFile(test_data_path / "dat/AldoJF250k_000000.dat") as f:
        header, data = f.read()

    n_frames = 53
    assert header.size == n_frames
    assert data.shape == (n_frames, 256, 1024)

    # Read reference data using numpy
    with np.load(test_data_path / "dat/AldoJF250k.npz") as f:
        ref_header = f["headers"]
        ref_data = f["frames"]

    # Check that the data is the same
    assert np.all(ref_header == header)
    assert np.all(ref_data == data)


@pytest.mark.withdata
def test_read_single_chip(test_data_path):

    # Read all frames from the .dat file
    with JungfrauDataFile(test_data_path / "dat/AldoJF65k_000000.dat") as f:
        header, data = f.read()

    n_frames = 113
    assert header.size == n_frames
    assert data.shape == (n_frames, 256, 256)

    # Read reference data using numpy
    with np.load(test_data_path / "dat/AldoJF65k.npz") as f:
        ref_header = f["headers"]
        ref_data = f["frames"]

    # Check that the data is the same
    assert np.all(ref_header == header)
    assert np.all(ref_data == data)


import pytest 
import numpy as np

import aare

def test_apply_calibration_small_data():
    # The raw data consists of 10 4x5 images
    raw = np.zeros((10, 4, 5), dtype=np.uint16)

    # We need a pedestal for each gain, so 3 
    pedestal = np.zeros((3, 4, 5), dtype=np.float32)

    # And the same for calibration
    calibration = np.ones((3, 4, 5), dtype=np.float32)

    # Set the known values, probing one pixel in each gain
    raw[0, 0, 0] = 100 #ADC value of 100, gain 0
    pedestal[0, 0, 0] = 10
    calibration[0, 0, 0] = 43.7

    raw[2, 3, 3] = (1<<14) + 1000 #ADC value of 1000, gain 1
    pedestal[1, 3, 3] = 500
    calibration[1, 3, 3] = 2.0

    raw[1,1,4] = (3<<14) + 857 #ADC value of 857, gain 2
    pedestal[2,1,4] = 100
    calibration[2,1,4] = 3.0

    

    data = aare.apply_calibration(raw, pd = pedestal, cal = calibration)


    # The formula that is applied is:
    # calibrated = (raw - pedestal) / calibration
    assert data.shape == (10, 4, 5)
    assert data[0, 0, 0] == (100 - 10) / 43.7
    assert data[2, 3, 3] ==  (1000 - 500) / 2.0
    assert data[1, 1, 4] == (857 - 100) / 3.0

    # Other pixels should be zero
    assert data[2,2,2] == 0
    assert data[0,1,1] == 0
    assert data[1,3,0] == 0


@pytest.fixture
def raw_data_3x2x2():
    raw = np.zeros((3, 2, 2), dtype=np.uint16)
    raw[0, 0, 0] = 100
    raw[1,0, 0] =  200
    raw[2, 0, 0] = 300

    raw[0, 0, 1] = (1<<14) + 100
    raw[1, 0, 1] = (1<<14) + 200
    raw[2, 0, 1] = (1<<14) + 300

    raw[0, 1, 0] = (1<<14) + 37
    raw[1, 1, 0] =  38
    raw[2, 1, 0] = (3<<14) + 39

    raw[0, 1, 1] = (3<<14) + 100
    raw[1, 1, 1] = (3<<14) + 200
    raw[2, 1, 1] = (3<<14) + 300
    return raw

def test_calculate_pedestal(raw_data_3x2x2):
    # Calculate the pedestal
    pd = aare.calculate_pedestal(raw_data_3x2x2)
    assert pd.shape == (3, 2, 2)
    assert pd.dtype == np.float64
    assert pd[0, 0, 0] == 200
    assert pd[1, 0, 0] == 0
    assert pd[2, 0, 0] == 0

    assert pd[0, 0, 1] == 0
    assert pd[1, 0, 1] == 200
    assert pd[2, 0, 1] == 0

    assert pd[0, 1, 0] == 38
    assert pd[1, 1, 0] == 37
    assert pd[2, 1, 0] == 39

    assert pd[0, 1, 1] == 0
    assert pd[1, 1, 1] == 0
    assert pd[2, 1, 1] == 200

def test_calculate_pedestal_float(raw_data_3x2x2):
    #results should be the same for float
    pd2 = aare.calculate_pedestal_float(raw_data_3x2x2)
    assert pd2.shape == (3, 2, 2)
    assert pd2.dtype == np.float32
    assert pd2[0, 0, 0] == 200
    assert pd2[1, 0, 0] == 0
    assert pd2[2, 0, 0] == 0

    assert pd2[0, 0, 1] == 0
    assert pd2[1, 0, 1] == 200
    assert pd2[2, 0, 1] == 0

    assert pd2[0, 1, 0] == 38
    assert pd2[1, 1, 0] == 37
    assert pd2[2, 1, 0] == 39

    assert pd2[0, 1, 1] == 0
    assert pd2[1, 1, 1] == 0
    assert pd2[2, 1, 1] == 200

def test_calculate_pedestal_g0(raw_data_3x2x2):
    pd = aare.calculate_pedestal_g0(raw_data_3x2x2)
    assert pd.shape == (2, 2)
    assert pd.dtype == np.float64
    assert pd[0, 0] == 200
    assert pd[1, 0] == 38
    assert pd[0, 1] == 0
    assert pd[1, 1] == 0

def test_calculate_pedestal_g0_float(raw_data_3x2x2):
    pd = aare.calculate_pedestal_g0_float(raw_data_3x2x2)
    assert pd.shape == (2, 2)
    assert pd.dtype == np.float32
    assert pd[0, 0] == 200
    assert pd[1, 0] == 38
    assert pd[0, 1] == 0
    assert pd[1, 1] == 0

def test_count_switching_pixels(raw_data_3x2x2):
    # Count the number of pixels that switched gain
    count = aare.count_switching_pixels(raw_data_3x2x2)
    assert count.shape == (2, 2)
    assert count.sum() == 8
    assert count[0, 0] == 0 
    assert count[1, 0] == 2 
    assert count[0, 1] == 3 
    assert count[1, 1] == 3


import pytest 
import numpy as np

from aare import _aare #import the C++ module
from conftest import test_data_path


def test_cluster_vector_can_be_converted_to_numpy():
    cv = _aare.ClusterVector_Cluster3x3i()
    arr = np.array(cv, copy=False)
    assert arr.shape == (0,)  # 4 for x, y, size, energy and 9 for the cluster data


def test_ClusterVector(): 
    """Test ClusterVector""" 

    clustervector = _aare.ClusterVector_Cluster3x3i()
    assert clustervector.cluster_size_x == 3
    assert clustervector.cluster_size_y == 3
    assert clustervector.item_size() == 4+9*4
    assert clustervector.frame_number == 0
    assert clustervector.size == 0

    cluster = _aare.Cluster3x3i(0,0,np.ones(9, dtype=np.int32))
    
    clustervector.push_back(cluster) 
    assert clustervector.size == 1

    with pytest.raises(TypeError):  # Or use the appropriate exception type
        clustervector.push_back(_aare.Cluster2x2i(0,0,np.ones(4, dtype=np.int32)))

    with pytest.raises(TypeError): 
        clustervector.push_back(_aare.Cluster3x3f(0,0,np.ones(9, dtype=np.float32))) 

def test_Interpolator(): 
    """Test Interpolator"""

    ebins = np.linspace(0,10, 20, dtype=np.float64)
    xbins = np.linspace(0, 5, 30, dtype=np.float64)
    ybins = np.linspace(0, 5, 30, dtype=np.float64)

    etacube = np.zeros(shape=[30, 30, 20], dtype=np.float64)
    interpolator = _aare.Interpolator(etacube, xbins, ybins, ebins)

    assert interpolator.get_ietax().shape == (30,30,20)
    assert interpolator.get_ietay().shape == (30,30,20)
    clustervector = _aare.ClusterVector_Cluster3x3i()
    
    cluster = _aare.Cluster3x3i(0,0, np.ones(9, dtype=np.int32))
    clustervector.push_back(cluster) 

    interpolated_photons = interpolator.interpolate(clustervector)

    assert interpolated_photons.size == 1

    assert interpolated_photons[0]["x"] == -1 
    assert interpolated_photons[0]["y"] == -1
    assert interpolated_photons[0]["energy"] == 4 #eta_sum = 4, dx, dy = -1,-1 m_ietax = 0, m_ietay = 0

    clustervector = _aare.ClusterVector_Cluster2x2i()

    cluster = _aare.Cluster2x2i(0,0, np.ones(4, dtype=np.int32))
    clustervector.push_back(cluster) 

    interpolated_photons = interpolator.interpolate(clustervector)

    assert interpolated_photons.size == 1

    assert interpolated_photons[0]["x"] == 0
    assert interpolated_photons[0]["y"] == 0
    assert interpolated_photons[0]["energy"] == 4



def test_calculate_eta(): 
    """Calculate Eta""" 
    clusters = _aare.ClusterVector_Cluster3x3i() 
    clusters.push_back(_aare.Cluster3x3i(0,0, np.ones(9, dtype=np.int32)))
    clusters.push_back(_aare.Cluster3x3i(0,0, np.array([1,1,1,2,2,2,3,3,3])))

    eta2 = _aare.calculate_eta2(clusters)

    assert eta2.shape == (2,2)
    assert eta2[0,0] == 0.5
    assert eta2[0,1] == 0.5
    assert eta2[1,0] == 0.5
    assert eta2[1,1] == 0.6 #1/5

def test_cluster_finder(): 
    """Test ClusterFinder""" 

    clusterfinder = _aare.ClusterFinder_Cluster3x3i([100,100])

    #frame = np.random.rand(100,100)
    frame = np.zeros(shape=[100,100])

    clusterfinder.find_clusters(frame)

    clusters = clusterfinder.steal_clusters(False) #conversion does not work

    assert clusters.size == 0












import pytest 
import numpy as np
import boost_histogram as bh
import time
from pathlib import Path
import pickle

from aare import ClusterFile
from aare import _aare
from conftest import test_data_path


def test_create_cluster_vector():
    cv = _aare.ClusterVector_Cluster3x3i()
    assert cv.cluster_size_x == 3
    assert cv.cluster_size_y == 3
    assert cv.size == 0


def test_push_back_on_cluster_vector():
    cv = _aare.ClusterVector_Cluster2x2i()
    assert cv.cluster_size_x == 2
    assert cv.cluster_size_y == 2
    assert cv.size == 0

    cluster = _aare.Cluster2x2i(19, 22, np.ones(4, dtype=np.int32))
    cv.push_back(cluster)
    assert cv.size == 1

    arr = np.array(cv, copy=False)
    assert arr[0]['x'] == 19
    assert arr[0]['y'] == 22


def test_make_a_hitmap_from_cluster_vector():
    cv = _aare.ClusterVector_Cluster3x3i()

    # Push back 4 clusters with different positions
    cv.push_back(_aare.Cluster3x3i(0, 0, np.ones(9, dtype=np.int32)))
    cv.push_back(_aare.Cluster3x3i(1, 1, np.ones(9, dtype=np.int32)))
    cv.push_back(_aare.Cluster3x3i(1, 1, np.ones(9, dtype=np.int32)))
    cv.push_back(_aare.Cluster3x3i(2, 2, np.ones(9, dtype=np.int32)))

    ref = np.zeros((5, 5), dtype=np.int32)
    ref[0,0] = 1
    ref[1,1] = 2
    ref[2,2] = 1


    img = _aare.hitmap((5,5), cv)
    # print(img)
    # print(ref)
    assert (img == ref).all()
    


# Make the compiled classes that live in _aare available from aare.
from . import _aare


from ._aare import File, RawMasterFile, RawSubFile, JungfrauDataFile
from ._aare import Pedestal_d, Pedestal_f, ClusterFinder_Cluster3x3i, VarClusterFinder
from ._aare import DetectorType
from ._aare import hitmap
from ._aare import ROI

# from ._aare import ClusterFinderMT, ClusterCollector, ClusterFileSink, ClusterVector_i

from .ClusterFinder import ClusterFinder, ClusterCollector, ClusterFinderMT, ClusterFileSink, ClusterFile
from .ClusterVector import ClusterVector


from ._aare import fit_gaus, fit_pol1, fit_scurve, fit_scurve2
from ._aare import Interpolator
from ._aare import calculate_eta2


from ._aare import apply_custom_weights

from .CtbRawFile import CtbRawFile
from .RawFile import RawFile
from .ScanParameters import ScanParameters

from .utils import random_pixels, random_pixel, flat_list, add_colorbar


#make functions available in the top level API
from .func import *

from .calibration import *
from ._aare import apply_calibration, count_switching_pixels
from ._aare import calculate_pedestal, calculate_pedestal_float, calculate_pedestal_g0, calculate_pedestal_g0_float

from ._aare import VarClusterFinder



import pytest
import numpy as np
from aare import RawSubFile, DetectorType


@pytest.mark.withdata
def test_read_a_jungfrau_RawSubFile(test_data_path):

    # Starting with f1 there is now 7 frames left in the series of files
    with RawSubFile(test_data_path / "raw/jungfrau/jungfrau_single_d0_f1_0.raw", DetectorType.Jungfrau, 512, 1024, 16) as f:
        assert f.frames_in_file == 7

        headers, frames = f.read()

    assert headers.size == 7
    assert frames.shape == (7, 512, 1024)
    

    for i,h in zip(range(4,11,1), headers):
        assert h["frameNumber"] == i

    # Compare to canned data using numpy
    data = np.load(test_data_path / "raw/jungfrau/jungfrau_single_0.npy")
    assert np.all(data[3:] == frames)

@pytest.mark.withdata
def test_iterate_over_a_jungfrau_RawSubFile(test_data_path):

    data = np.load(test_data_path / "raw/jungfrau/jungfrau_single_0.npy")

    # Given the first subfile in a series we can read all frames from f0, f1, f2...fN
    with RawSubFile(test_data_path / "raw/jungfrau/jungfrau_single_d0_f0_0.raw", DetectorType.Jungfrau, 512, 1024, 16) as f:
        i = 0
        for header, frame in f:
            assert header["frameNumber"] == i+1
            assert np.all(frame == data[i])
            i += 1
        assert i == 10
        assert header["frameNumber"] == 10



import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable

def random_pixels(n_pixels, xmin=0, xmax=512, ymin=0, ymax=1024):
    """Return a list of random pixels.
    
    Args:
        n_pixels (int): Number of pixels to return.
        rows (int): Number of rows in the image.
        cols (int): Number of columns in the image.
        
    Returns:
        list: List of (row, col) tuples.
    """
    return [(np.random.randint(ymin, ymax), np.random.randint(xmin, xmax)) for _ in range(n_pixels)]


def random_pixel(xmin=0, xmax=512, ymin=0, ymax=1024):
    """Return a random pixel.
    
    Returns:
        tuple: (row, col)
    """
    return random_pixels(1, xmin, xmax, ymin, ymax)[0]

def flat_list(xss):
    """Flatten a list of lists."""
    return [x for xs in xss for x in xs]

def add_colorbar(ax, im, size="5%", pad=0.05):
    """Add a colorbar with the same height as the image."""
    divider = make_axes_locatable(ax)
    cax = divider.append_axes("right", size=size, pad=pad)
    plt.colorbar(im, cax=cax)
    return ax, im, cax


# SPDX-License-Identifier: LGPL-3.0-or-other
# Copyright (C) 2021 Contributors to the Aare Package
"""
Script to update VERSION file with semantic versioning if provided as an argument, or with 0.0.0 if no argument is provided.
"""

import sys
import os
import re
from datetime import datetime

from packaging.version import Version, InvalidVersion


SCRIPT_DIR = os.path.dirname(os.path.abspath(__file__))

def is_integer(value): 
    try:
        int(value)
    except ValueError:
        return False
    else:
        return True


def get_version():

    # Check at least one argument is passed
    if len(sys.argv) < 2:
        version =   datetime.today().strftime('%Y.%-m.%-d')
    else: 
        version = sys.argv[1]

    try:
        v = Version(version)  # normalize check if version follows PEP 440 specification
        
        version_normalized = version.replace("-", ".")

        version_normalized =  re.sub(r'0*(\d+)', lambda m : str(int(m.group(0))), version_normalized) #remove leading zeros 

        return version_normalized 

    except InvalidVersion as e:
        print(f"Invalid version {version}. Version format must follow semantic versioning format of python PEP 440 version identification specification.")
        sys.exit(1)
    

def write_version_to_file(version):
    version_file_path = os.path.join(SCRIPT_DIR, "VERSION")
    with open(version_file_path, "w") as version_file:
        version_file.write(version)
    print(f"Version {version} written to VERSION file.")

# Main script
if __name__ == "__main__":

    version = get_version()
    write_version_to_file(version)



import matplotlib.pyplot as plt
import numpy as np
from aare import fit_gaus, fit_pol1
from aare import gaus, pol1

textpm = f"±"  #
textmu = f"μ"  #
textsigma = f"σ"  #



# ================================= Gauss fit =================================
# Parameters
mu = np.random.uniform(1, 100)  # Mean of Gaussian
sigma = np.random.uniform(4, 20)  # Standard deviation
num_points = 10000  # Number of points for smooth distribution
noise_sigma = 100

# Generate Gaussian distribution
data = np.random.normal(mu, sigma, num_points)

# Generate errors for each point
errors = np.abs(np.random.normal(0, sigma, num_points))  # Errors with mean 0, std 0.5

# Create subplot
fig0, ax0 = plt.subplots(1, 1, num=0, figsize=(12, 8))

x = np.histogram(data, bins=30)[1][:-1] + 0.05
y = np.histogram(data, bins=30)[0]
yerr = errors[:30]


# Add the errors as error bars in the step plot
ax0.errorbar(x, y, yerr=yerr, fmt=". ", capsize=5)
ax0.grid()

par, err = fit_gaus(x, y, yerr)
print(par, err)

x = np.linspace(x[0], x[-1], 1000)
ax0.plot(x, gaus(x, par), marker="")
ax0.set(xlabel="x", ylabel="Counts", title=f"A0 = {par[0]:0.2f}{textpm}{err[0]:0.2f}\n"
                                           f"{textmu} = {par[1]:0.2f}{textpm}{err[1]:0.2f}\n"
                                           f"{textsigma} = {par[2]:0.2f}{textpm}{err[2]:0.2f}\n"
                                           f"(init: {textmu}: {mu:0.2f}, {textsigma}: {sigma:0.2f})")
fig0.tight_layout()



# ================================= pol1 fit =================================
# Parameters
n_points = 40

# Generate random slope and intercept (origin)
slope = np.random.uniform(-10, 10)  # Random slope between 0.5 and 2.0
intercept = np.random.uniform(-10, 10)  # Random intercept between -10 and 10

# Generate random x values
x_values = np.random.uniform(-10, 10, n_points)

# Calculate y values based on the linear function y = mx + b + error
errors = np.abs(np.random.normal(0, np.random.uniform(1, 5), n_points))
var_points = np.random.normal(0, np.random.uniform(0.1, 2), n_points)
y_values = slope * x_values + intercept + var_points

fig1, ax1 = plt.subplots(1, 1, num=1, figsize=(12, 8))
ax1.errorbar(x_values, y_values, yerr=errors, fmt=". ", capsize=5)
par, err = fit_pol1(x_values, y_values, errors)


x = np.linspace(np.min(x_values), np.max(x_values), 1000)
ax1.plot(x, pol1(x, par), marker="")
ax1.set(xlabel="x", ylabel="y", title=f"a = {par[0]:0.2f}{textpm}{err[0]:0.2f}\n"
                                      f"b = {par[1]:0.2f}{textpm}{err[1]:0.2f}\n"
                                      f"(init: {slope:0.2f}, {intercept:0.2f})")
fig1.tight_layout()

plt.show()






from ._aare import ClusterVector_Cluster3x3i
import numpy as np

def ClusterVector(cluster_size, dtype = np.int32):

    if dtype == np.int32 and cluster_size == (3,3):
        return ClusterVector_Cluster3x3i()
    else:
        raise ValueError(f"Unsupported dtype: {dtype}. Only np.int32 is supported.")



#Calibration related functions
import numpy as np
def load_calibration(fname, hg0=False):
    """
    Load calibration data from a file.
    
    Parameters:
    fname (str): Path to the calibration file.
    hg0 (bool): If True, load HG0 calibration data instead of G0.

    """
    gains = 3
    rows = 512
    cols = 1024
    with open(fname, 'rb') as f:
        cal = np.fromfile(f, count=gains * rows * cols, dtype=np.double).reshape(
            gains, rows, cols
        )
        if hg0:
            cal[0] = np.fromfile(f, count=rows * cols, dtype=np.double).reshape(rows, cols)
    return cal


from . import _aare
import numpy as np

_supported_cluster_sizes = [(2,2), (3,3), (5,5), (7,7), (9,9),]

def _type_to_char(dtype):
    if dtype == np.int32:
        return 'i'
    elif dtype == np.float32:
        return 'f'
    elif dtype == np.float64:
        return 'd'
    else:
        raise ValueError(f"Unsupported dtype: {dtype}. Only np.int32, np.float32, and np.float64 are supported.")

def _get_class(name, cluster_size, dtype):
    """
    Helper function to get the class based on the name, cluster size, and dtype.
    """
    try:
        class_name = f"{name}_Cluster{cluster_size[0]}x{cluster_size[1]}{_type_to_char(dtype)}"
        cls = getattr(_aare, class_name)
    except AttributeError:
        raise ValueError(f"Unsupported combination of type and cluster size: {dtype}/{cluster_size} when requesting {class_name}")
    return cls



def ClusterFinder(image_size, cluster_size, n_sigma=5, dtype = np.int32, capacity = 1024):
    """
    Factory function to create a ClusterFinder object. Provides a cleaner syntax for 
    the templated ClusterFinder in C++.
    """
    cls = _get_class("ClusterFinder", cluster_size, dtype)
    return cls(image_size, n_sigma=n_sigma, capacity=capacity)



def ClusterFinderMT(image_size, cluster_size = (3,3), dtype=np.int32, n_sigma=5, capacity = 1024, n_threads = 3): 
    """ 
    Factory function to create a ClusterFinderMT object. Provides a cleaner syntax for 
    the templated ClusterFinderMT in C++.
    """

    cls = _get_class("ClusterFinderMT", cluster_size, dtype)
    return cls(image_size, n_sigma=n_sigma, capacity=capacity, n_threads=n_threads)



def ClusterCollector(clusterfindermt, cluster_size = (3,3), dtype=np.int32): 
    """ 
    Factory function to create a ClusterCollector object. Provides a cleaner syntax for 
    the templated ClusterCollector in C++.
    """

    cls = _get_class("ClusterCollector", cluster_size, dtype)
    return cls(clusterfindermt)

def ClusterFileSink(clusterfindermt, cluster_file, dtype=np.int32): 
    """ 
    Factory function to create a ClusterCollector object. Provides a cleaner syntax for 
    the templated ClusterCollector in C++.
    """

    cls = _get_class("ClusterFileSink", clusterfindermt.cluster_size, dtype)
    return cls(clusterfindermt, cluster_file)


def ClusterFile(fname, cluster_size=(3,3), dtype=np.int32, chunk_size = 1000):
    """
    Factory function to create a ClusterFile object. Provides a cleaner syntax for
    the templated ClusterFile in C++.

    .. code-block:: python

        from aare import ClusterFile
        
        with ClusterFile("clusters.clust", cluster_size=(3,3), dtype=np.int32) as cf:
            # cf is now a ClusterFile_Cluster3x3i object but you don't need to know that.
            for clusters in cf:
                # Loop over clusters in chunks of 1000 
                # The type of clusters will be a ClusterVector_Cluster3x3i in this case

    """

    cls = _get_class("ClusterFile", cluster_size, dtype)
    return cls(fname, chunk_size=chunk_size)



from . import _aare
import numpy as np
from .ScanParameters import ScanParameters

class RawFile(_aare.RawFile):
    def __init__(self, fname, chunk_size = 1):
        super().__init__(fname)
        self._chunk_size = chunk_size


    def read(self) -> tuple:
        """Read the entire file.
        Seeks to the beginning of the file before reading.

        Returns:
            tuple: header, data
        """
        self.seek(0)
        return self.read_n(self.total_frames)

    @property
    def scan_parameters(self):
        """Return the scan parameters.

        Returns:
            ScanParameters: Scan parameters.
        """
        return ScanParameters(self.master.scan_parameters)
    
    @property
    def master(self):
        """Return the master file.

        Returns:
            RawMasterFile: Master file.
        """
        return super().master
    
    def __len__(self) -> int:
        """Return the number of frames in the file.

        Returns:
            int: Number of frames in file.
        """
        return super().frames_in_file

    def __enter__(self):
        return self
    
    def __exit__(self, exc_type, exc_value, traceback):
        pass

    def __iter__(self):
        return self
    
    def __next__(self):
        try:
            if self._chunk_size == 1:
                return self.read_frame()
            else:
                return self.read_n(self._chunk_size)
            
                
        except RuntimeError:
            # TODO! find a good way to check that we actually have the right exception
            raise StopIteration



import numpy as np
from . import _aare


class AdcSar04Transform64to16:
    def __call__(self, data):
        return _aare.adc_sar_04_decode64to16(data)

class AdcSar05Transform64to16:
    def __call__(self, data):
        return _aare.adc_sar_05_decode64to16(data)

class Moench05Transform:
    #Could be moved to C++ without changing the interface
    def __init__(self):
        self.pixel_map = _aare.GenerateMoench05PixelMap()

    def __call__(self, data):
        return np.take(data.view(np.uint16), self.pixel_map)
    

class Moench05Transform1g:
    #Could be moved to C++ without changing the interface
    def __init__(self):
        self.pixel_map = _aare.GenerateMoench05PixelMap1g()

    def __call__(self, data):
        return np.take(data.view(np.uint16), self.pixel_map)
    

class Moench05TransformOld:
    #Could be moved to C++ without changing the interface
    def __init__(self):
        self.pixel_map = _aare.GenerateMoench05PixelMapOld()

    def __call__(self, data):
        return np.take(data.view(np.uint16), self.pixel_map)


class Matterhorn02Transform:
    def __init__(self):
        self.pixel_map = _aare.GenerateMH02FourCounterPixelMap()

    def __call__(self, data):
        counters = int(data.size / 48**2 / 2)
        if counters == 1:
            return np.take(data.view(np.uint16), self.pixel_map[0])
        else:
            return np.take(data.view(np.uint16), self.pixel_map[0:counters])


#on import generate the pixel maps to avoid doing it every time
moench05 = Moench05Transform()
moench05_1g = Moench05Transform1g()
moench05_old = Moench05TransformOld()
matterhorn02 = Matterhorn02Transform()
adc_sar_04_64to16 = AdcSar04Transform64to16()
adc_sar_05_64to16 = AdcSar05Transform64to16()